# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Client to handle communication with SciCat servers."""

from __future__ import annotations

import dataclasses
import datetime
import re
import warnings
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Tuple, Union
from urllib.parse import quote_plus

import requests

from . import model
from ._base_model import convert_download_to_user_model
from .dataset import Dataset
from .error import ScicatCommError, ScicatLoginError
from .file import File
from .filesystem import RemotePath
from .logging import get_logger
from .pid import PID
from .typing import DownloadConnection, FileTransfer, UploadConnection
from .util.credentials import SecretStr, StrStorage


class Client:
    """SciCat client to communicate with a server.

    Clients hold all information needed to communicate with a SciCat instance
    and a filesystem that holds data files (via ``file_transfer``).

    Use :func:`Client.from_token` or :func:`Client.from_credentials` to initialize
    a client instead of the constructor directly.

    See the user guide for typical usage patterns.
    In particular, `Downloading Datasets <../../user-guide/downloading.ipynb>`_
    and `Uploading Datasets <../../user-guide/uploading.ipynb>`_.
    """

    def __init__(
        self,
        *,
        client: ScicatClient,
        file_transfer: Optional[FileTransfer],
    ):
        """Initialize a client.

        Do not use directly, instead use :func:`Client.from_token`
        or :func:`Client.from_credentials`!
        """
        self._client = client
        self._file_transfer = file_transfer

    @classmethod
    def from_token(
        cls,
        *,
        url: str,
        token: Union[str, StrStorage],
        file_transfer: Optional[FileTransfer] = None,
    ) -> Client:
        """Create a new client and authenticate with a token.

        Parameters
        ----------
        url:
            URL of the SciCat api.
        token:
            User token to authenticate with SciCat.
        file_transfer:
            Handler for down-/uploads of files.

        Returns
        -------
        :
            A new client.
        """
        return Client(
            client=ScicatClient.from_token(url=url, token=token),
            file_transfer=file_transfer,
        )

    # TODO rename to login? and provide logout?
    @classmethod
    def from_credentials(
        cls,
        *,
        url: str,
        username: Union[str, StrStorage],
        password: Union[str, StrStorage],
        file_transfer: Optional[FileTransfer] = None,
    ) -> Client:
        """Create a new client and authenticate with username and password.

        Parameters
        ----------
        url:
            URL of the SciCat api.
            It should include the suffix `api/vn` where `n` is a number.
        username:
            Name of the user.
        password:
            Password of the user.
        file_transfer:
            Handler for down-/uploads of files.

        Returns
        -------
        :
            A new client.
        """
        return Client(
            client=ScicatClient.from_credentials(
                url=url, username=username, password=password
            ),
            file_transfer=file_transfer,
        )

    @classmethod
    def without_login(
        cls, *, url: str, file_transfer: Optional[FileTransfer] = None
    ) -> Client:
        """Create a new client without authentication.

        The client can only download public datasets and not upload at all.

        Parameters
        ----------
        url:
            URL of the SciCat api.
            It should include the suffix `api/vn` where `n` is a number.
        file_transfer:
            Handler for down-/uploads of files.

        Returns
        -------
        :
            A new client.
        """
        return Client(
            client=ScicatClient.without_login(url=url), file_transfer=file_transfer
        )

    @property
    def scicat(self) -> ScicatClient:
        """Low-level client for SciCat.

        Should typically not be used by users of Scitacean!
        """
        return self._client

    @property
    def file_transfer(self) -> Optional[FileTransfer]:
        """Stored handler for file down-/uploads."""
        return self._file_transfer

    def get_dataset(
        self,
        pid: Union[str, PID],
        strict_validation: bool = False,
        attachments: bool = False,
    ) -> Dataset:
        """Download a dataset from SciCat.

        Does not download any files.

        Parameters
        ----------
        pid:
            ID of the dataset. Must include the prefix, i.e. have the form
            ``prefix/dataset-id``.
        strict_validation:
            If ``True``, the dataset must pass validation.
            If ``False``, a dataset is still returned if validation fails.
            Note that some dataset fields may have a bad value or type.
            A warning will be logged if validation fails.
        attachments:
            Select whether to download attachments.
            If this is ``False``, the attachments of the returned dataset are ``None``.
            They can be downloaded later using :meth:`Client.download_attachments_for`.

        Returns
        -------
        :
            A new dataset.
        """
        pid = PID.parse(pid)
        dataset = self.scicat.get_dataset_model(
            pid, strict_validation=strict_validation
        )

        try:
            orig_datablocks = self.scicat.get_orig_datablocks(
                pid, strict_validation=strict_validation
            )
        except ScicatCommError:
            # TODO more precise error handling. We only want to set to None if
            #   communication succeeded and the dataset exists but there simply
            #   are no datablocks.
            orig_datablocks = None

        attachment_models = (
            self.scicat.get_attachments_for_dataset(pid) if attachments else None
        )

        return Dataset.from_download_models(
            dataset_model=dataset,
            orig_datablock_models=orig_datablocks or [],
            attachment_models=attachment_models,
        )

    def upload_new_dataset_now(self, dataset: Dataset) -> Dataset:
        """Upload a dataset as a new entry to SciCat immediately.

        The dataset is inserted as a new entry in the database and will
        never overwrite existing data.
        To this end, the input dataset's ID is ignored and a new one is
        assigned automatically.

        Attachments are also uploaded automatically.
        This happens after the upload of files and the dataset itself.
        So if uploading the attachments fails, check the dataset in SciCat to
        determine which attachments you need to re-upload
        (using :meth:`ScicatClient.create_attachment_for_dataset`).

        Parameters
        ----------
        dataset:
            The dataset to upload.

        Returns
        -------
        :
            A copy of the input dataset with fields adjusted
            according to the response of the server.

        Raises
        ------
        scitacean.ScicatCommError
            If the upload to SciCat fails.
        RuntimeError
            If the file upload fails or if a critical error is encountered
            and some files or a partial dataset are left on the servers.
            Note the error message if that happens.
        """
        dataset = dataset.replace(source_folder=self._source_folder_for(dataset))
        files_to_upload = _files_to_upload(dataset.files)
        self.scicat.validate_dataset_model(dataset.make_upload_model())
        with self._connect_for_file_upload(dataset, files_to_upload) as con:
            # TODO check if any remote file is out of date.
            #  if so, raise an error. We never overwrite remote files!
            uploaded_files = con.upload_files(*files_to_upload)
            dataset = dataset.replace_files(*uploaded_files)
            try:
                finalized_model = self.scicat.create_dataset_model(
                    dataset.make_upload_model()
                )
            except ScicatCommError:
                con.revert_upload(*uploaded_files)
                raise

        with_new_pid = dataset.replace(_read_only={"pid": finalized_model.pid})
        finalized_orig_datablocks = self._upload_orig_datablocks(
            with_new_pid.make_datablock_upload_models().orig_datablocks
        )
        finalized_attachments = self._upload_attachments_for_dataset(
            with_new_pid.make_attachment_upload_models(),
            dataset_id=with_new_pid.pid,  # type: ignore[arg-type]
        )

        return Dataset.from_download_models(
            dataset_model=finalized_model,
            orig_datablock_models=finalized_orig_datablocks,
            attachment_models=finalized_attachments,
        )

    def _upload_orig_datablocks(
        self, orig_datablocks: Optional[List[model.UploadOrigDatablock]]
    ) -> List[model.DownloadOrigDatablock]:
        if not orig_datablocks:
            return []

        try:
            return [
                self.scicat.create_orig_datablock(orig_datablock)
                for orig_datablock in orig_datablocks
            ]
        except ScicatCommError as exc:
            raise RuntimeError(
                "Failed to upload original datablocks for SciCat dataset "
                f"{orig_datablocks[0].datasetId}:"
                f"\n{exc.args}\nThe dataset and data files were successfully uploaded "
                "but are not linked with each other. Please fix the dataset manually!"
            ) from exc

    def _upload_attachments_for_dataset(
        self, attachments: List[model.UploadAttachment], *, dataset_id: PID
    ) -> List[model.DownloadAttachment]:
        try:
            return [
                self.scicat.create_attachment_for_dataset(
                    attachment, dataset_id=dataset_id
                )
                for attachment in attachments
            ]
        except ScicatCommError as exc:
            raise RuntimeError(
                f"Failed to upload attachments for SciCat dataset {dataset_id}:"
                f"\n{exc.args}\nThe dataset and data files were successfully uploaded "
                "and will not be reverted. Please upload the attachments manually!"
            ) from exc

    @contextmanager
    def _connect_for_file_upload(
        self, dataset: Dataset, files_to_upload: Iterable[File]
    ) -> Iterator[UploadConnection]:
        if not files_to_upload:
            yield _NullUploadConnection()
        else:
            with self._expect_file_transfer().connect_for_upload(dataset) as con:
                yield con

    def _source_folder_for(self, dataset: Dataset) -> RemotePath:
        if self._file_transfer is not None:
            return self._file_transfer.source_folder_for(dataset)
        if dataset.source_folder is None:
            raise ValueError(
                "Cannot determine source_folder for dataset because "
                "the dataset's source_folder is None and there is no file transfer."
            )
        return dataset.source_folder

    def _expect_file_transfer(self) -> FileTransfer:
        if self.file_transfer is None:
            raise ValueError(
                "Cannot upload/download files because no file transfer is set. "
                "Specify one when constructing a client."
            )
        return self.file_transfer

    def download_files(
        self,
        dataset: Dataset,
        *,
        target: Union[str, Path],
        select: FileSelector = True,
        checksum_algorithm: Optional[str] = None,
        force: bool = False,
    ) -> Dataset:
        r"""Download files of a dataset.

        Makes selected files available on the local filesystem using the file transfer
        object stored in the client.

        Parameters
        ----------
        dataset:
            Download files of this dataset.
        target:
            Files are stored to this path on the local filesystem.
        select:
            A file ``f`` is downloaded if ``select`` is

            - **True**: downloaded
            - **False**: not downloaded
            - A **string**: if ``f.remote_path`` equals this string
            - A **list[str]** or **tuple[str]**: if ``f.remote_path``
              equals any of these strings
            - An **re.Pattern** as returned by :func:`re.compile`:
              if this pattern matches ``f.remote_path`` using :func:`re.search`
            - A **Callable[File]**: if this callable returns ``True`` for ``f``
        checksum_algorithm:
            Select an algorithm for computing file checksums.
            The data block normally determines the algorithm, but this argument can
            override the stored value if the data block is broken
            or no algorithm has been set.
        force:
            If ``True``, download files regardless of whether they already exist
            locally.
            This bypasses the checksum computation of pre-existing local files.

        Returns
        -------
        :
            A copy of the input dataset with files replaced to reflect the downloads.
            That is, the ``local_path`` of all downloaded files is set.

        Examples
        --------
        Download all files

        .. code-block:: python

            client.download_files(dataset, target="./data", select=True)

        Download a specific file

        .. code-block:: python

            client.download_files(dataset, target="./data", select="my-file.dat")

        Download all files with a ``nxs`` extension

        .. code-block:: python

            import re
            client.download_files(
                dataset,
                target="./data",
                select=re.compile(r"\.nxs$")
            )
            # or
            client.download_files(
                dataset,
                target="./data",
                select=lambda file: file.remote_path.suffix == ".nxs"
            )
        """
        if dataset.source_folder is None:
            raise ValueError("Dataset has no source folder, cannot download files.")
        target = Path(target)
        # TODO undo if later fails but only if no files were written
        target.mkdir(parents=True, exist_ok=True)
        files = _select_files(select, dataset)
        downloaded_files = [
            f.downloaded(local_path=target / f.remote_path.to_local()) for f in files
        ]
        if not force:
            to_download = _remove_up_to_date_local_files(
                downloaded_files, checksum_algorithm=checksum_algorithm
            )
        else:
            to_download = downloaded_files

        if not to_download:
            return dataset.replace_files(*downloaded_files)

        with self._connect_for_file_download() as con:
            con.download_files(
                remote=[
                    p
                    for f in to_download
                    if (p := f.remote_access_path(dataset.source_folder)) is not None
                ],
                local=[f.local_path for f in to_download],  # type: ignore[misc]
            )
        for f in to_download:
            f.validate_after_download()
        return dataset.replace_files(*downloaded_files)

    @contextmanager
    def _connect_for_file_download(self) -> Iterator[DownloadConnection]:
        if self.file_transfer is None:
            raise ValueError(
                "Cannot download files because no file transfer is set. "
                "Specify one when constructing a client."
            )
        with self.file_transfer.connect_for_download() as con:
            yield con

    def download_attachments_for(self, target: Dataset) -> Dataset:
        """Download all attachments for a given object.

        The target object must have an ID, that is, it must represent an entry
        in SciCat and not just a locally created object.

        If the input already has attachments, they will be overwritten and a
        ``UserWarning`` will be raised.

        Parameters
        ----------
        target:
            Download attachments for this object.

        Returns
        -------
        :
            A copy of the input dataset with attachments replaced
            with the downloaded models.
        """
        if target.pid is None:
            raise ValueError(
                "Cannot download attachments because the dataset has no PID."
            )
        if target.attachments is not None:
            warnings.warn(
                "Downloading attachments for a dataset that already has "
                "attachments. The existing attachments will be overwritten.",
                stacklevel=2,
            )
        return target.replace(
            attachments=convert_download_to_user_model(
                self.scicat.get_attachments_for_dataset(target.pid)
            )
        )


class ScicatClient:
    """Low-level client to call the SciCat API."""

    def __init__(
        self,
        url: str,
        token: Optional[Union[str, StrStorage]],
        timeout: Optional[datetime.timedelta],
    ):
        # Need to add a final /
        self._base_url = url[:-1] if url.endswith("/") else url
        self._timeout = datetime.timedelta(seconds=10) if timeout is None else timeout
        self._token: Optional[StrStorage] = (
            SecretStr(token) if isinstance(token, str) else token
        )

    @classmethod
    def from_token(
        cls,
        url: str,
        token: Union[str, StrStorage],
        timeout: Optional[datetime.timedelta] = None,
    ) -> ScicatClient:
        """Create a new low-level client and authenticate with a token.

        Parameters
        ----------
        url:
            URL of the SciCat api.
        token:
            User token to authenticate with SciCat.
        timeout:
            Timeout for all API requests.

        Returns
        -------
        :
            A new low-level client.
        """
        return ScicatClient(url=url, token=token, timeout=timeout)

    @classmethod
    def from_credentials(
        cls,
        url: str,
        username: Union[str, StrStorage],
        password: Union[str, StrStorage],
        timeout: Optional[datetime.timedelta] = None,
    ) -> ScicatClient:
        """Create a new low-level client and authenticate with username and password.

        Parameters
        ----------
        url:
            URL of the SciCat api.
            It should include the suffix `api/vn` where `n` is a number.
        username:
            Name of the user.
        password:
            Password of the user.
        timeout:
            Timeout for all API requests.

        Returns
        -------
        :
            A new low-level client.
        """
        if not isinstance(username, StrStorage):
            username = SecretStr(username)
        if not isinstance(password, StrStorage):
            password = SecretStr(password)
        return ScicatClient(
            url=url,
            token=SecretStr(
                _get_token(
                    url=url,
                    username=username,
                    password=password,
                    timeout=timeout if timeout else datetime.timedelta(seconds=10),
                )
            ),
            timeout=timeout,
        )

    @classmethod
    def without_login(
        cls, url: str, timeout: Optional[datetime.timedelta] = None
    ) -> ScicatClient:
        """Create a new low-level client without authentication.

        The client can only download public datasets and not upload at all.

        Parameters
        ----------
        url:
            URL of the SciCat api.
            It should include the suffix `api/vn` where `n` is a number.
        timeout:
            Timeout for all API requests.

        Returns
        -------
        :
            A new low-level client.
        """
        return ScicatClient(url=url, token=None, timeout=timeout)

    def get_dataset_model(
        self, pid: PID, strict_validation: bool = False
    ) -> model.DownloadDataset:
        """Fetch a dataset from SciCat.

        Parameters
        ----------
        pid:
            Unique ID of the dataset.
            Must include the facility ID.
        strict_validation:
            If ``True``, the dataset must pass validation.
            If ``False``, a dataset is still returned if validation fails.
            Note that some dataset fields may have a bad value or type.
            A warning will be logged if validation fails.

        Returns
        -------
        :
            A model of the dataset.

        Raises
        ------
        scitacean.ScicatCommError
            If the dataset does not exist or communication fails for some other reason.
        """
        dset_json = self._call_endpoint(
            cmd="get",
            url=f"datasets/{quote_plus(str(pid))}",
            operation="get_dataset_model",
        )
        if not dset_json:
            raise ScicatCommError(
                f"Cannot get dataset with {pid=}, "
                f"no such dataset in SciCat at {self._base_url}."
            )
        return model.construct(
            model.DownloadDataset,
            _strict_validation=strict_validation,
            **dset_json,
        )

    def get_orig_datablocks(
        self, pid: PID, strict_validation: bool = False
    ) -> List[model.DownloadOrigDatablock]:
        """Fetch all orig datablocks from SciCat for a given dataset.

        Parameters
        ----------
        pid:
            Unique ID of the *dataset*.
            Must include the facility ID.
        strict_validation:
            If ``True``, the datablocks must pass validation.
            If ``False``, datablocks are still returned if validation fails.
            Note that some fields may have a bad value or type.
            A warning will be logged if validation fails.

        Returns
        -------
        :
            Models of the orig datablocks.

        Raises
        ------
        scitacean.ScicatCommError
            If communication fails.
        """
        dblock_json = self._call_endpoint(
            cmd="get",
            url=f"datasets/{quote_plus(str(pid))}/origdatablocks",
            operation="get_orig_datablocks",
        )
        return [
            _make_orig_datablock(dblock, strict_validation=strict_validation)
            for dblock in dblock_json
        ]

    def get_attachments_for_dataset(
        self, pid: PID, strict_validation: bool = False
    ) -> List[model.DownloadAttachment]:
        """Fetch all attachments from SciCat for a given dataset.

        Parameters
        ----------
        pid:
            Unique ID of the *dataset*.
            Must include the facility ID.
        strict_validation:
            If ``True``, the attachments must pass validation.
            If ``False``, attachments are still returned if validation fails.
            Note that some fields may have a bad value or type.
            A warning will be logged if validation fails.

        Returns
        -------
        :
            Models of the attachments.

        Raises
        ------
        scitacean.ScicatCommError
            If communication fails.
        """
        attachment_json = self._call_endpoint(
            cmd="get",
            url=f"datasets/{quote_plus(str(pid))}/attachments",
            operation="get_attachments_for_dataset",
        )
        return [
            model.construct(
                model.DownloadAttachment,
                _strict_validation=strict_validation,
                **attachment,
            )
            for attachment in attachment_json
        ]

    def create_dataset_model(
        self, dset: Union[model.UploadDerivedDataset, model.UploadRawDataset]
    ) -> model.DownloadDataset:
        """Create a new dataset in SciCat.

        The dataset PID must be either

        - ``None``, in which case SciCat assigns an ID,
        - an unused id, in which case SciCat uses it for the new dataset.

        If the ID already exists, creation will fail without
        modification to the database.
        Unless the new dataset is identical to the existing one,
        in which case, nothing happens.

        Parameters
        ----------
        dset:
            Model of the dataset to create.

        Returns
        -------
        :
            The uploaded dataset as returned by SciCat.
            This is the input plus some modifications.

        Raises
        ------
        scitacean.ScicatCommError
            If SciCat refuses the dataset or communication
            fails for some other reason.
        """
        uploaded = self._call_endpoint(
            cmd="post", url="datasets", data=dset, operation="create_dataset_model"
        )
        return model.construct(
            model.DownloadDataset, _strict_validation=False, **uploaded
        )

    def create_orig_datablock(
        self, dblock: model.UploadOrigDatablock
    ) -> model.DownloadOrigDatablock:
        """Create a new orig datablock in SciCat.

        The datablock PID must be either

        - ``None``, in which case SciCat assigns an ID.
        - An unused id, in which case SciCat uses it for the new datablock.

        If the ID already exists, creation will fail without
        modification to the database.

        Parameters
        ----------
        dblock:
            Model of the orig datablock to create.

        Raises
        ------
        scitacean.ScicatCommError
            If SciCat refuses the datablock or communication
            fails for some other reason.
        """
        uploaded = self._call_endpoint(
            cmd="post",
            url="origdatablocks",
            data=dblock,
            operation="create_orig_datablock",
        )
        return model.construct(
            model.DownloadOrigDatablock, _strict_validation=False, **uploaded
        )

    def create_attachment_for_dataset(
        self,
        attachment: model.UploadAttachment,
        *,
        dataset_id: PID,
    ) -> model.DownloadAttachment:
        """Create a new attachment for a dataset in SciCat.

        The attachment ID must be either

        - ``None``, in which case SciCat assigns an ID.
        - An unused id, in which case SciCat uses it for the new attachment.

        If the ID already exists, creation will fail without
        modification to the database.

        Parameters
        ----------
        attachment:
            Model of the attachment to create.
        dataset_id:
            ID of the dataset to which the attachment belongs.

        Raises
        ------
        scitacean.ScicatCommError
            If SciCat refuses the attachment or communication
            fails for some other reason.
        """
        uploaded = self._call_endpoint(
            cmd="post",
            url=f"datasets/{quote_plus(str(dataset_id))}/attachments",
            data=attachment,
            operation="create_attachment",
        )
        if not uploaded:
            raise ScicatCommError(
                f"Failed to upload attachment for dataset with pid={dataset_id}. "
                "The server reported success but did not return a finalized attachment."
                " This likely means that there is no dataset with this ID."
            )
        return model.construct(
            model.DownloadAttachment, _strict_validation=False, **uploaded
        )

    def validate_dataset_model(
        self, dset: Union[model.UploadDerivedDataset, model.UploadRawDataset]
    ) -> None:
        """Validate a dataset in SciCat.

        Parameters
        ----------
        dset:
            Model of the dataset to validate.

        Raises
        ------
        ValueError
            If the dataset does not pass validation.
        """
        response = self._call_endpoint(
            cmd="post",
            url="datasets/isValid",
            data=dset,
            operation="validate_dataset_model",
        )
        if not response["valid"]:
            raise ValueError(f"Dataset {dset} did not pass validation in SciCat.")

    def _send_to_scicat(
        self, *, cmd: str, url: str, data: Optional[model.BaseModel] = None
    ) -> requests.Response:
        if self._token is not None:
            token = self._token.get_str()
            params = {"access_token": token}
            headers = {"Authorization": f"Bearer {token}"}
        else:
            token = ""
            params = {}
            headers = {}

        if data is not None:
            headers["Content-Type"] = "application/json"

        try:
            return requests.request(
                method=cmd,
                url=url,
                data=data.model_dump_json(exclude_none=True)
                if data is not None
                else None,
                params=params,
                headers=headers,
                timeout=self._timeout.seconds,
                stream=False,
                verify=True,
            )
        except Exception as exc:
            # Remove concrete request function call from backtrace to hide the token.
            # Also modify the error message to strip out the token.
            # It shows up, e.g. in urllib3.exceptions.NewConnectionError.
            # This turns the exception args into strings.
            # But we have little use of more structured errors, so that should be fine.
            raise type(exc)(
                tuple(_strip_token(arg, token) for arg in exc.args)
            ) from None

    def _call_endpoint(
        self,
        *,
        cmd: str,
        url: str,
        data: Optional[model.BaseModel] = None,
        operation: str,
    ) -> Any:
        full_url = _url_concat(self._base_url, url)
        logger = get_logger()
        logger.info("Calling SciCat API at %s for operation '%s'", full_url, operation)

        response = self._send_to_scicat(cmd=cmd, url=full_url, data=data)
        if not response.ok:
            logger.error(
                "SciCat API call to %s failed: %s %s: %s",
                full_url,
                response.status_code,
                response.reason,
                response.text,
            )
            raise ScicatCommError(
                f"Error in operation '{operation}': {response.status_code} "
                f"{response.reason}: {response.text}"
            )
        logger.info("API call successful for operation '%s'", operation)

        return None if not response.text else response.json()


def _url_concat(a: str, b: str) -> str:
    # Combine two pieces or a URL without handling absolute
    # paths as in urljoin.
    a = a if a.endswith("/") else (a + "/")
    b = b[1:] if b.endswith("/") else b
    return a + b


def _strip_token(error: Any, token: str) -> str:
    err = str(error)
    err = re.sub(r"token=[\w\-./]+", "token=<HIDDEN>", err)
    if token:  # token can be ""
        err = err.replace(token, "<HIDDEN>")
    return err


def _make_orig_datablock(
    fields: Dict[str, Any], strict_validation: bool
) -> model.DownloadOrigDatablock:
    files = [
        model.construct(
            model.DownloadDataFile, _strict_validation=strict_validation, **file_fields
        )
        for file_fields in fields["dataFileList"]
    ]
    return model.construct(
        model.DownloadOrigDatablock,
        _strict_validation=strict_validation,
        **{**fields, "dataFileList": files},
    )


def _log_in_via_users_login(
    url: str, username: StrStorage, password: StrStorage, timeout: datetime.timedelta
) -> requests.Response:
    # Currently only used for functional accounts.
    response = requests.post(
        _url_concat(url, "Users/login"),
        json={"username": username.get_str(), "password": password.get_str()},
        stream=False,
        verify=True,
        timeout=timeout.seconds,
    )
    if not response.ok:
        get_logger().info(
            "Failed to log in via endpoint Users/login: %s", response.json()["error"]
        )
    return response


def _log_in_via_auth_msad(
    url: str, username: StrStorage, password: StrStorage, timeout: datetime.timedelta
) -> requests.Response:
    # Used for user accounts.
    import re

    # Strip the api/vn suffix
    base_url = re.sub(r"/api/v\d+/?", "", url)
    response = requests.post(
        _url_concat(base_url, "auth/msad"),
        json={"username": username.get_str(), "password": password.get_str()},
        stream=False,
        verify=True,
        timeout=timeout.seconds,
    )
    if not response.ok:
        get_logger().error(
            "Failed to log in via auth/msad: %s", response.json()["error"]
        )
    return response


def _get_token(
    url: str, username: StrStorage, password: StrStorage, timeout: datetime.timedelta
) -> str:
    """Log in using the provided username + password.

    Returns a token for the given user.
    """
    # Users/login only works for functional accounts and auth/msad for regular users.
    # Try both and see what works. This is not nice but seems to be the only
    # feasible solution right now.
    get_logger().info("Logging in to %s", url)

    response = _log_in_via_users_login(
        url=url, username=username, password=password, timeout=timeout
    )
    if response.ok:
        return str(response.json()["id"])  # not sure if semantically correct

    response = _log_in_via_auth_msad(
        url=url, username=username, password=password, timeout=timeout
    )
    if response.ok:
        return str(response.json()["access_token"])

    get_logger().error("Failed log in:  %s", response.json()["error"])
    raise ScicatLoginError(response.content)


FileSelector = Union[
    bool, str, List[str], Tuple[str], re.Pattern, Callable[[File], bool]
]


def _file_selector(select: FileSelector) -> Callable[[File], bool]:
    if select is True:
        return lambda _: True
    if select is False:
        return lambda _: False
    if isinstance(select, str):
        return lambda f: f.remote_path == select
    if isinstance(select, (list, tuple)):
        return lambda f: f.remote_path in select  # type: ignore[operator]
    if isinstance(select, re.Pattern):
        return lambda f: (
            select.search(f.remote_path.posix) is not None  # type: ignore[union-attr]
        )
    return select


def _select_files(select: FileSelector, dataset: Dataset) -> List[File]:
    selector = _file_selector(select)
    return [f for f in dataset.files if selector(f)]


def _remove_up_to_date_local_files(
    files: List[File], checksum_algorithm: Optional[str]
) -> List[File]:
    def is_up_to_date(file: File) -> bool:
        if checksum_algorithm is not None:
            file = dataclasses.replace(file, checksum_algorithm=checksum_algorithm)
        return file.local_is_up_to_date()

    return [
        file
        for file in files
        if not (
            file.local_path.exists() and is_up_to_date(file)  # type: ignore[union-attr]
        )
    ]


def _files_to_upload(files: Iterable[File]) -> List[File]:
    for file in files:
        if file.is_on_local and file.is_on_remote:
            raise ValueError(
                f"Refusing to upload file at remote_path={file.remote_path} "
                "because it is both on local and remote and it is unclear what "
                "to do. If you want to perform the upload, set the local path to None."
            )
    return [file for file in files if file.local_path is not None]


class _NullUploadConnection:
    """File-upload connection that does not upload anything.

    Raises if called with files because uit should only be used by Client
    when there are no files to upload.
    """

    def upload_files(self, *files: File) -> List[File]:
        """Raise if given files."""
        if files:
            raise RuntimeError("Internal error: Bad upload connection")
        return []

    def revert_upload(self, *files: File) -> None:
        """Raise if given files."""
        if files:
            raise RuntimeError("Internal error: Bad upload connection")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Python version-independent dataclasses."""

import dataclasses
from typing import Any, Callable, Type, TypeVar

T = TypeVar("T")


try:
    from typing import dataclass_transform
except ImportError:
    from typing import Tuple, Union

    F = TypeVar("F")

    def dataclass_transform(
        *,
        eq_default: bool = True,
        order_default: bool = False,
        kw_only_default: bool = False,
        frozen_default: bool = False,
        field_specifiers: Tuple[Union[Type[Any], Callable[..., Any]], ...] = (),
        **kwargs: Any,
    ) -> Callable[[T], T]:
        def impl(f: F) -> F:
            return f

        return impl


@dataclass_transform()
def dataclass_optional_args(
    kw_only: bool = False, slots: bool = False, **kwargs: Any
) -> Callable[[Type[T]], Type[T]]:
    """Create a dataclass with modern arguments."""
    try:
        # Python 3.10+
        return dataclasses.dataclass(kw_only=kw_only, slots=slots, **kwargs)
    except TypeError:
        # Fallback for older Python
        return dataclasses.dataclass(**kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""String-formatting tools."""

from string import Formatter
from typing import Any, Iterable, Optional, Tuple

from ..filesystem import escape_path

try:
    from typing import LiteralString, TypeVar

    StrOrLiteralStr = TypeVar("StrOrLiteralStr", LiteralString, str)
    del LiteralString, TypeVar
except ImportError:
    StrOrLiteralStr = str  # type: ignore[misc]


class DatasetPathFormatter(Formatter):
    """Formatter that inserts dataset fields and escapes paths.

    This formatter automatically modifies format strings such that the
    following two are equivalent (up to escaping, see below).

    .. code-block:: python

        formatter = DatasetPathFormatter()
        formatter.format("{owner}-{pid.pid}-{uid}", dset=dset, uid=uid)
        # is equivalent to (up to escaping)
        "{dset.owner}-{dset.pid.pid}-{uid}".format(dset=dset, uid=uid)

    This means that all format fields are transformed to access attributes
    of the keyword argument ``dset``.
    Except for ``"uid"`` which is preserved to allow for a dedicated ``uid`` keyword.

    In addition, all field values are escaped as filesystem paths
    using :func:`scitacean.filesystem.escape_path`:

    .. code-block:: python

        formatter = DatasetPathFormatter()
        formatter.format("a string {owner}", dset=dset)
        # is equivalent to
        "a string {}".format(escape_path("{dset.owner}".format(dset=dset)))

    Note that only formatted fields are escaped, not the result as a whole.

    Fields that are used by the formatter must not be ``None``.
    Otherwise, a :class:`ValueError` will be raised.
    """

    def parse(
        self, format_string: str
    ) -> Iterable[
        Tuple[
            StrOrLiteralStr,
            Optional[StrOrLiteralStr],
            Optional[StrOrLiteralStr],
            Optional[StrOrLiteralStr],
        ]
    ]:
        """Parse a format string."""

        def add0(field_name: Optional[str]) -> Optional[str]:
            if field_name == "uid":
                return field_name
            if isinstance(field_name, str):
                return "dset." + field_name
            return None

        return ((t[0], add0(t[1]), t[2], t[3]) for t in super().parse(format_string))

    def format_field(self, value: Any, format_spec: str) -> str:
        """Format a field with the given spec."""
        if value is None:
            raise ValueError("Cannot format path, dataset field is None")
        formatted: str = super().format_field(value, format_spec)
        return escape_path(formatted)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
# mypy: disable-error-code="no-untyped-def, return-value, arg-type, union-attr"

import dataclasses
import tempfile
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Iterator

import paramiko
import pytest

from scitacean import Dataset, File, RemotePath
from scitacean.testing.client import FakeClient
from scitacean.testing.sftp import IgnorePolicy, skip_if_not_sftp
from scitacean.transfer.sftp import (
    SFTPDownloadConnection,
    SFTPFileTransfer,
    SFTPUploadConnection,
)


@pytest.fixture(scope="session", autouse=True)
def server(request, sftp_fileserver):
    skip_if_not_sftp(request)


def test_download_one_file(sftp_access, sftp_connect_with_username_password, tmp_path):
    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_download() as con:
        con.download_files(
            remote=[RemotePath("/data/seed/text.txt")], local=[tmp_path / "text.txt"]
        )
    assert (
        tmp_path.joinpath("text.txt").read_text() == "This is some text for testing.\n"
    )


def test_download_two_files(sftp_access, sftp_connect_with_username_password, tmp_path):
    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_download() as con:
        con.download_files(
            remote=[
                RemotePath("/data/seed/table.csv"),
                RemotePath("/data/seed/text.txt"),
            ],
            local=[tmp_path / "local-table.csv", tmp_path / "text.txt"],
        )
    assert tmp_path.joinpath("local-table.csv").read_text() == "7,2\n5,2\n"
    assert (
        tmp_path.joinpath("text.txt").read_text() == "This is some text for testing.\n"
    )


def test_upload_one_file_source_folder_in_dataset(
    sftp_access, sftp_connect_with_username_password, tmp_path, sftp_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/upload"))
    tmp_path.joinpath("file0.txt").write_text("File to test upload123")

    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_upload(dataset=ds) as con:
        assert con.source_folder == RemotePath("/data/upload")
        con.upload_files(
            File.from_local(path=tmp_path / "file0.txt", remote_path="upload_0.txt")
        )

    assert (
        sftp_data_dir.joinpath("upload", "upload_0.txt").read_text()
        == "File to test upload123"
    )


def test_upload_one_file_source_folder_in_transfer(
    sftp_access, sftp_connect_with_username_password, tmp_path, sftp_data_dir
):
    ds = Dataset(type="raw", owner="librarian")
    tmp_path.joinpath("file1.txt").write_text("File no. 2")

    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        source_folder="/data/upload/{owner}",
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_upload(dataset=ds) as con:
        assert con.source_folder == RemotePath("/data/upload/librarian")
        con.upload_files(
            File.from_local(
                path=tmp_path / "file1.txt", remote_path=RemotePath("upload_1.txt")
            )
        )

    assert (
        sftp_data_dir.joinpath("upload", "librarian", "upload_1.txt").read_text()
        == "File no. 2"
    )


def test_upload_two_files(
    sftp_access, sftp_connect_with_username_password, tmp_path, sftp_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/upload2"))
    tmp_path.joinpath("file2.1.md").write_text("First part of file 2")
    tmp_path.joinpath("file2.2.md").write_text("Second part of file 2")

    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_upload(dataset=ds) as con:
        assert con.source_folder == RemotePath("/data/upload2")
        con.upload_files(
            File.from_local(path=tmp_path / "file2.1.md", base_path=tmp_path),
            File.from_local(path=tmp_path / "file2.2.md", base_path=tmp_path),
        )

    assert (
        sftp_data_dir.joinpath("upload2", "file2.1.md").read_text()
        == "First part of file 2"
    )
    assert (
        sftp_data_dir.joinpath("upload2", "file2.2.md").read_text()
        == "Second part of file 2"
    )


def test_upload_one_file_existing_source_folder(
    sftp_access, sftp_connect_with_username_password, tmp_path, sftp_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/upload-multiple"))
    tmp_path.joinpath("file3.1.md").write_text("First part of file 3")
    tmp_path.joinpath("file3.2.md").write_text("Second part of file 3")

    # First upload to ensure the folder exists.
    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_upload(dataset=ds) as con:
        assert con.source_folder == RemotePath("/data/upload-multiple")
        con.upload_files(
            File.from_local(path=tmp_path / "file3.1.md", base_path=tmp_path),
        )

    # Second upload to test uploading to existing folder.
    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_upload(dataset=ds) as con:
        assert con.source_folder == RemotePath("/data/upload-multiple")
        con.upload_files(
            File.from_local(path=tmp_path / "file3.2.md", base_path=tmp_path),
        )

    assert (
        sftp_data_dir.joinpath("upload-multiple", "file3.1.md").read_text()
        == "First part of file 3"
    )
    assert (
        sftp_data_dir.joinpath("upload-multiple", "file3.2.md").read_text()
        == "Second part of file 3"
    )


def test_revert_all_uploaded_files_single(
    sftp_access, sftp_connect_with_username_password, tmp_path, sftp_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/revert-all-test-1"))
    tmp_path.joinpath("file3.txt").write_text("File that should get reverted")

    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_upload(dataset=ds) as con:
        file = File.from_local(path=tmp_path / "file3.txt", base_path=tmp_path)
        con.upload_files(file)
        con.revert_upload(file)

    assert "revert-all-test-1" not in (p.name for p in sftp_data_dir.iterdir())


def test_revert_all_uploaded_files_two(
    sftp_access, sftp_connect_with_username_password, tmp_path, sftp_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/revert-all-test-2"))
    tmp_path.joinpath("file3.1.txt").write_text("File that should get reverted 1")
    tmp_path.joinpath("file3.2.txt").write_text("File that should get reverted 2")

    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_upload(dataset=ds) as con:
        file1 = File.from_local(path=tmp_path / "file3.1.txt", base_path=tmp_path)
        file2 = File.from_local(path=tmp_path / "file3.2.txt", base_path=tmp_path)
        con.upload_files(file1, file2)
        con.revert_upload(file1, file2)

    assert "revert-all-test-2" not in (p.name for p in sftp_data_dir.iterdir())


def test_revert_one_uploaded_file(
    sftp_access, sftp_connect_with_username_password, tmp_path, sftp_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/revert-test"))
    tmp_path.joinpath("file4.txt").write_text("File that should get reverted")
    tmp_path.joinpath("file5.txt").write_text("File that should be kept")

    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_upload(dataset=ds) as con:
        file4 = File.from_local(path=tmp_path / "file4.txt", base_path=tmp_path)
        file5 = File.from_local(path=tmp_path / "file5.txt", base_path=tmp_path)
        con.upload_files(file4, file5)
        con.revert_upload(file4)

    assert "file4.txt" not in (
        p.name for p in (sftp_data_dir / "revert-test").iterdir()
    )
    assert (
        sftp_data_dir.joinpath("revert-test", "file5.txt").read_text()
        == "File that should be kept"
    )


def test_stat_uploaded_file(
    sftp_access, sftp_connect_with_username_password, tmp_path, sftp_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/upload6"))
    tmp_path.joinpath("file6.txt").write_text("File to test upload no 6")

    sftp = SFTPFileTransfer(
        host=sftp_access.host,
        port=sftp_access.port,
        connect=sftp_connect_with_username_password,
    )
    with sftp.connect_for_upload(dataset=ds) as con:
        [uploaded] = con.upload_files(
            File.from_local(path=tmp_path / "file6.txt", remote_path="upload_6.txt")
        )

    st = (sftp_data_dir / "upload6" / "upload_6.txt").stat()
    assert uploaded.size == st.st_size

    # Set in docker-compose
    assert uploaded.remote_uid == "1000"
    assert uploaded.remote_gid == "1000"

    uploaded = dataclasses.replace(uploaded, local_path=None)
    assert datetime.now(tz=timezone.utc) - uploaded.creation_time < timedelta(seconds=5)


class CorruptingSFTP(paramiko.SFTPClient):
    """Appends bytes to uploaded files to simulate a broken transfer."""

    def put(self, localpath, remotepath, callback=None, confirm=True):
        with open(localpath, "r") as f:
            content = f.read()
        with tempfile.TemporaryDirectory() as tempdir:
            corrupted_path = Path(tempdir) / "corrupted"
            with open(corrupted_path, "w") as f:
                f.write(content + "\nevil bytes")
            super().put(str(corrupted_path), remotepath, callback, confirm)


class CorruptingTransfer(paramiko.Transport):
    """Uses CorruptingSFTP to simulate a broken connection."""

    def open_sftp_client(self) -> paramiko.SFTPClient:
        return CorruptingSFTP.from_transport(self)


@pytest.fixture()
def sftp_corrupting_connect(sftp_access, sftp_connection_config):
    def connect(host: str, port: int) -> paramiko.SFTPClient:
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(IgnorePolicy())
        client.connect(
            hostname=host,
            port=port,
            username=sftp_access.user.username,
            password=sftp_access.user.password,
            allow_agent=False,
            look_for_keys=False,
            transport_factory=CorruptingTransfer,
        )
        return client.open_sftp()

    return connect


class RaisingSFTP(paramiko.SFTPClient):
    def put(self, localpath, remotepath, callback=None, confirm=True):
        raise RuntimeError("Upload disabled")


class RaisingTransfer(paramiko.Transport):
    def open_sftp_client(self) -> paramiko.SFTPClient:
        return RaisingSFTP.from_transport(self)


@pytest.fixture()
def sftp_raising_connect(sftp_access):
    def connect(host: str, port: int) -> paramiko.SFTPClient:
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(IgnorePolicy())
        client.connect(
            hostname=host,
            port=port,
            username=sftp_access.user.username,
            password=sftp_access.user.password,
            allow_agent=False,
            look_for_keys=False,
            transport_factory=RaisingTransfer,
        )
        return client.open_sftp()

    return connect


def test_upload_file_reverts_if_upload_fails(
    sftp_access, sftp_raising_connect, tmp_path, sftp_data_dir
):
    ds = Dataset(type="raw", source_folder=RemotePath("/data/upload8"))
    tmp_path.joinpath("file8.txt").write_text("File to test upload no 8")

    sftp = SFTPFileTransfer(
        host=sftp_access.host, port=sftp_access.port, connect=sftp_raising_connect
    )
    with sftp.connect_for_upload(dataset=ds) as con:
        with pytest.raises(RuntimeError):
            con.upload_files(
                File.from_local(
                    path=tmp_path / "file8.txt",
                    remote_path=RemotePath("upload_8.txt"),
                )
            )

    assert "upload8" not in (p.name for p in sftp_data_dir.iterdir())


class SFTPTestFileTransfer(SFTPFileTransfer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @contextmanager
    def connect_for_download(self) -> Iterator[SFTPDownloadConnection]:
        with super().connect_for_download() as connection:
            yield connection

    @contextmanager
    def connect_for_upload(self, dataset: Dataset) -> Iterator[SFTPUploadConnection]:
        with super().connect_for_upload(dataset=dataset) as connection:
            yield connection


# This test is referenced in the docs.
def test_client_with_sftp(
    require_sftp_fileserver,
    sftp_access,
    sftp_connect_with_username_password,
    sftp_data_dir,
    tmp_path,
):
    tmp_path.joinpath("file1.txt").write_text("File contents")

    client = FakeClient.without_login(
        url="",
        file_transfer=SFTPTestFileTransfer(
            connect=sftp_connect_with_username_password,
            host=sftp_access.host,
            port=sftp_access.port,
        ),
    )
    ds = Dataset(
        access_groups=["group1"],
        contact_email="p.stibbons@uu.am",
        creation_location="UU",
        creation_time=datetime(2023, 6, 23, 10, 0, 0),
        owner="PonderStibbons",
        owner_group="uu",
        principal_investigator="MustrumRidcully",
        source_folder="/data",
        type="raw",
    )
    ds.add_local_files(tmp_path / "file1.txt", base_path=tmp_path)
    finalized = client.upload_new_dataset_now(ds)

    downloaded = client.get_dataset(finalized.pid)
    downloaded = client.download_files(downloaded, target=tmp_path / "download")

    assert sftp_data_dir.joinpath("file1.txt").read_text() == "File contents"
    assert downloaded.files[0].local_path.read_text() == "File contents"



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

"""Custom warning classes."""


class VisibleDeprecationWarning(UserWarning):
    """Visible deprecation warning.

    By default, Python and in particular Jupyter will not show deprecation
    warnings, so this class can be used when a very visible warning is helpful.
    """


VisibleDeprecationWarning.__module__ = "scitacean"



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Common functions for HTML reprs."""

import html
from datetime import datetime
from typing import Any, List, Optional

from .._internal.dataclass_wrapper import dataclass_optional_args
from ..filesystem import RemotePath
from ..model import History, Lifecycle, Relationship, Technique
from ..pid import PID
from . import _resources


@dataclass_optional_args(kw_only=True, frozen=True, slots=True)
class Field:
    name: str
    value: Any
    type: type
    description: str
    read_only: bool
    required: bool
    error: Optional[str]
    main: bool


_TYPE_NAME = {
    str: "str",
    int: "int",
    bool: "bool",
    dict: "dict",
    datetime: "datetime",
    History: "History",
    Lifecycle: "Lifecycle",
    PID: "PID",
    RemotePath: "RemotePath",
    List[str]: "list[str]",
    List[PID]: "list[PID]",
    List[Relationship]: "list[Relationship]",
    List[Technique]: "list[Technique]",
    List[dict]: "list[dict]",  # type: ignore[type-arg]
}


def format_value(val: Any) -> str:
    """Return a str repr for a field value."""
    if val is None:
        return '<span class="cean-empty-field">None</span>'
    if isinstance(val, datetime):
        return val.strftime("%Y-%m-%d %H:%M:%S%z")
    return html.escape(str(val))


def format_type(typ: Any) -> str:
    """Return a concise str repr for a type."""
    # Types are strings model.py because of
    # from __future__ import annotations
    if isinstance(typ, str) and typ.startswith("Optional["):
        typ = typ[9:-1]
    try:
        return _TYPE_NAME[typ]
    except KeyError:
        return html.escape(str(typ))


def format_field_flag(field: Field) -> str:
    """Return HTML markup for field flags."""
    flags = ""
    if field.read_only:
        flags += f'<div class="cean-lock">{_resources.image("lock.svg")}</div>'
    if field.required:
        flags += '<div style="color: var(--jp-error-color0)">*</div>'
    return flags



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

"""High-level interface for SciCat."""

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from .client import Client
from .datablock import OrigDatablock
from .dataset import Dataset
from .error import FileUploadError, IntegrityError, ScicatCommError, ScicatLoginError
from .file import File
from .filesystem import RemotePath
from .model import Attachment, DatasetType
from .pid import PID
from .thumbnail import Thumbnail
from .warning import VisibleDeprecationWarning

__all__ = (
    "Attachment",
    "Client",
    "Dataset",
    "DatasetType",
    "File",
    "FileUploadError",
    "IntegrityError",
    "OrigDatablock",
    "PID",
    "RemotePath",
    "ScicatCommError",
    "ScicatLoginError",
    "Thumbnail",
    "VisibleDeprecationWarning",
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
import importlib.resources
import time
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Iterable, Tuple, Union

import paramiko
import yaml


@dataclass
class SSHUser:
    username: str
    password: str


@dataclass
class SSHAccess:
    host: str
    port: int
    user: SSHUser


def _read_yaml(filename: str) -> Any:
    if hasattr(importlib.resources, "files"):
        # Use new API added in Python 3.9
        return yaml.safe_load(
            importlib.resources.files("scitacean.testing.ssh")
            .joinpath(filename)
            .read_text()
        )
    # Old API, deprecated as of Python 3.11
    return yaml.safe_load(
        importlib.resources.read_text("scitacean.testing.ssh", filename)
    )


@lru_cache(maxsize=1)
def _docker_compose_file() -> Dict[str, Any]:
    return _read_yaml("docker-compose-ssh-server.yaml")  # type: ignore[no-any-return]


def _seed_files() -> Iterable[Tuple[str, str]]:
    if hasattr(importlib.resources, "files"):
        # Use new API added in Python 3.9
        yield from (
            (file.name, file.read_text())
            for file in importlib.resources.files("scitacean.testing.ssh")
            .joinpath("ssh_server_seed")
            .iterdir()
        )
    else:
        # Old API, deprecated as of Python 3.11
        with importlib.resources.path(
            "scitacean.testing.ssh", "ssh_server_seed"
        ) as seed_dir:
            for path in seed_dir.iterdir():
                yield path.name, path.read_text()


def local_access() -> SSHAccess:
    config = _docker_compose_file()
    service = config["services"]["scitacean-test-ssh-server"]
    env = {k: v for k, v in map(lambda s: s.split("="), service["environment"])}
    return SSHAccess(
        host="localhost",
        port=service["ports"][0].split(":")[0],
        user=SSHUser(
            username=env["USER_NAME"],
            password=env["USER_PASSWORD"],
        ),
    )


def _copy_seed(target_seed_dir: Path) -> None:
    for name, content in _seed_files():
        target_seed_dir.joinpath(name).write_text(content)


def configure(target_dir: Union[Path, str]) -> Path:
    """Generate a config file for docker compose and copy seed data."""
    target_dir = Path(target_dir)
    target_seed_dir = target_dir / "data" / "seed"
    target_seed_dir.mkdir(parents=True)
    _copy_seed(target_seed_dir)

    config_target = target_dir / "docker-compose.yaml"
    config_target.write_text(yaml.dump(_docker_compose_file()))

    target_dir.joinpath(".env").write_text(
        f"""DATA_DIR={target_dir / 'data'}
SEED_DIR={target_seed_dir}"""
    )

    return config_target


def can_connect(ssh_access: SSHAccess) -> bool:
    try:
        _make_client(ssh_access)
    except paramiko.SSHException:
        return False
    return True


def wait_until_ssh_server_is_live(
    ssh_access: SSHAccess, max_time: float, n_tries: int
) -> None:
    # The container takes a while to be fully live.
    for _ in range(n_tries):
        if can_connect(ssh_access):
            return
        time.sleep(max_time / n_tries)
    if not can_connect(ssh_access):
        raise RuntimeError("Cannot connect to SSH server")


def cleanup_data_dir(
    ssh_access: SSHAccess, ssh_connect_with_username_password: Any
) -> None:
    # Delete all directories created by tests.
    # These are owned by root on the host and cannot be deleted by Python's tempfile.
    connection = ssh_connect_with_username_password(
        host=ssh_access.host, port=ssh_access.port
    )
    connection.run(
        "find /data -not -path '/data' -not -path '/data/seed' | xargs rm -rf",
        hide=True,
        in_stream=False,
    )


def _make_client(ssh_access: SSHAccess) -> paramiko.SSHClient:
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(IgnorePolicy())
    client.connect(
        hostname=ssh_access.host,
        port=ssh_access.port,
        username=ssh_access.user.username,
        password=ssh_access.user.password,
        allow_agent=False,
        look_for_keys=False,
    )
    return client


# Every time we create a container, it gets a new host key.
# So simply accept any host keys.
class IgnorePolicy(paramiko.MissingHostKeyPolicy):
    def missing_host_key(self, client: Any, hostname: Any, key: Any) -> None:
        return



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Main dataset structure."""

from __future__ import annotations

import dataclasses
import itertools
from datetime import datetime, timezone
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
)

from ._base_model import convert_download_to_user_model, convert_user_to_upload_model
from ._dataset_fields import DatasetBase
from .datablock import OrigDatablock
from .file import File
from .model import (
    Attachment,
    DatasetType,
    DownloadAttachment,
    DownloadDataset,
    DownloadOrigDatablock,
    UploadAttachment,
    UploadDerivedDataset,
    UploadOrigDatablock,
    UploadRawDataset,
)
from .pid import PID


class Dataset(DatasetBase):
    """Metadata and linked data files for a measurement, simulation, or analysis."""

    @classmethod
    def from_download_models(
        cls,
        dataset_model: DownloadDataset,
        orig_datablock_models: List[DownloadOrigDatablock],
        attachment_models: Optional[Iterable[DownloadAttachment]] = None,
    ) -> Dataset:
        """Construct a new dataset from SciCat download models.

        Parameters
        ----------
        dataset_model:
            Model of the dataset.
        orig_datablock_models:
            List of all associated original datablock models for the dataset.
        attachment_models:
            List of all associated attachment models for the dataset.
            Use ``None`` if the attachments were not downloaded.
            Use an empty list if the attachments were downloaded, but there aren't any.

        Returns
        -------
        :
            A new Dataset instance.
        """
        init_args, read_only = DatasetBase._prepare_fields_from_download(dataset_model)
        dset = cls(**init_args)
        for key, val in read_only.items():
            setattr(dset, key, val)
        dset._attachments = convert_download_to_user_model(  # type: ignore[assignment]
            attachment_models
        )
        if orig_datablock_models is not None:
            dset._orig_datablocks.extend(
                map(OrigDatablock.from_download_model, orig_datablock_models)
            )
        return dset

    @classmethod
    def fields(
        cls,
        dataset_type: Optional[Union[DatasetType, Literal["derived", "raw"]]] = None,
        read_only: Optional[bool] = None,
    ) -> Generator[Dataset.Field, None, None]:
        """Iterate over dataset fields.

        This is similar to :func:`dataclasses.fields`.

        Parameters
        ----------
        dataset_type:
            If set, return only the fields for this dataset type.
            If unset, do not filter fields.
        read_only:
            If true or false, return only fields which are read-only
            or allow write-access, respectively.
            If unset, do not filter fields.

        Returns
        -------
        :
            Iterable over the fields of datasets.
        """
        it = iter(DatasetBase._FIELD_SPEC)
        if dataset_type is not None:
            attr = (
                "used_by_derived"
                if dataset_type == DatasetType.DERIVED
                else "used_by_raw"
            )
            it = filter(lambda field: getattr(field, attr), it)
        if read_only is not None:
            it = filter(lambda field: field.read_only == read_only, it)
        yield from it

    def __str__(self) -> str:
        args = ", ".join(
            f"{field.name}={value}"
            for field in self.fields()
            if (value := getattr(self, field.name)) is not None
        )
        return f"Dataset({args})"

    def _repr_html_(self) -> str:
        # Import here to prevent cycle.
        from ._html_repr import dataset_html_repr

        return dataset_html_repr(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Dataset):
            return False
        eq = all(
            getattr(self, field.name) == getattr(other, field.name)
            for field in Dataset.fields()
        )
        eq = (
            eq
            and self._orig_datablocks == other._orig_datablocks
            and self._attachments == other._attachments
        )
        return eq

    @property
    def number_of_files(self) -> int:
        """Number of files in directly accessible storage in the dataset.

        This includes files on both the local and remote filesystems.

        Corresponds to OrigDatablocks.
        """
        return sum(len(tuple(dblock.files)) for dblock in self._orig_datablocks)

    @property
    def number_of_files_archived(self) -> int:
        """Total number of archived files in the dataset.

        Corresponds to Datablocks.
        """
        return 0

    @property
    def packed_size(self) -> int:
        """Total size of all datablock package files created for this dataset."""
        return 0

    @property
    def size(self) -> int:
        """Total size of files in directly accessible storage in the dataset.

        This includes files on both the local and remote filesystems.

        Corresponds to OrigDatablocks.
        """
        return sum(file.size for file in self.files)

    @property
    def files(self) -> Tuple[File, ...]:
        """Files linked with the dataset."""
        return tuple(
            itertools.chain.from_iterable(
                dblock.files for dblock in self._orig_datablocks
            )
        )

    @property
    def attachments(self) -> Optional[List[Attachment]]:
        """List of attachments for this dataset.

        This property can be in two distinct 'falsy' states:

        - ``dset.attachments is None``: It is unknown whether there are attachments.
          This happens when datasets are downloaded without downloading the attachments.
        - ``dset.attachments == []``: It is known that there are no attachments.
          This happens either when downloading datasets or when initializing datasets
          locally without assigning attachments.
        """
        return self._attachments

    @attachments.setter
    def attachments(self, attachments: Optional[List[Attachment]]) -> None:
        """List of attachments for this dataset.

        See the docs of the getter for an explanation of ``None`` vs ``[]``.
        It is unlikely that you ever need to set the attachments to ``None``.
        """
        self._attachments = attachments

    def add_files(self, *files: File, datablock: Union[int, str, PID] = -1) -> None:
        """Add files to the dataset."""
        self._get_or_add_orig_datablock(datablock).add_files(*files)

    def add_local_files(
        self,
        *paths: Union[str, Path],
        base_path: Union[str, Path] = "",
        datablock: Union[int, str] = -1,
    ) -> None:
        """Add files on the local file system to the dataset.

        Parameters
        ----------
        paths:
            Local paths to the files.
        base_path:
            The remote paths will be set up according to

            ``remotes = [path.relative_to(base_path) for path in paths]``.
        datablock:
            Advanced feature, do not set unless you know what this is!

            Select the orig datablock to store the file in.
            If an ``int``, use the datablock with that index.
            If a ``str`` or ``PID``, use the datablock with that id;
            if there is none with matching id, raise ``KeyError``.
        """
        self.add_files(
            *(File.from_local(path, base_path=base_path) for path in paths),
            datablock=datablock,
        )

    def replace(
        self,
        *,
        _read_only: Optional[Dict[str, Any]] = None,
        _orig_datablocks: Optional[List[OrigDatablock]] = None,
        **replacements: Any,
    ) -> Dataset:
        """Return a new dataset with replaced fields.

        Parameters starting with an underscore are for internal use.
        Using them may result in a broken dataset.

        Parameters
        ----------
        replacements:
            New field values.

        Returns
        -------
        :
            The new dataset has the same fields as the input but all fields given
            as keyword arguments are replaced by the given values.
        """
        _read_only = _read_only or {}

        def get_val(source: Dict[str, Any], name: str) -> Any:
            try:
                return source.pop(name)
            except KeyError:
                return getattr(self, name)

        read_only = {
            field.name: get_val(_read_only, field.name)
            for field in Dataset.fields(read_only=True)
        }
        DatasetBase._convert_readonly_fields_in_place(read_only)
        kwargs = {
            **{
                field.name: get_val(replacements, field.name)
                for field in Dataset.fields(read_only=False)
            },
        }
        attachments = get_val(replacements, "attachments")
        meta = get_val(replacements, "meta")

        if replacements or _read_only:
            raise TypeError(
                f"Invalid arguments: {', '.join((*replacements, *_read_only))}"
            )
        dset = Dataset(
            **kwargs,
            meta=meta,
        )
        dset._orig_datablocks.extend(
            _orig_datablocks if _orig_datablocks is not None else self._orig_datablocks
        )
        dset._attachments = list(attachments) if attachments is not None else None
        for key, val in read_only.items():
            setattr(dset, "_" + key, val)
        return dset

    def as_new(self) -> Dataset:
        """Return a new dataset with lifecycle-related fields erased.

        The returned dataset has the same fields as ``self``.
        But fields that indicate when the dataset was created or
        by who are set to ``None``.
        This if, for example, ``created_at``, ``history``, and ``lifecycle``.

        Returns
        -------
        :
            A new dataset without lifecycle-related fields.
        """
        return self.replace(
            _read_only={field.name: None for field in Dataset.fields(read_only=True)},
            creation_time=datetime.now(tz=timezone.utc),
        )

    def derive(
        self,
        *,
        keep: Iterable[str] = (
            "contact_email",
            "investigator",
            "orcid_of_owner",
            "owner",
            "owner_email",
            "techniques",
        ),
    ) -> Dataset:
        """Return a new dataset that is derived from self.

        The returned dataset has most fields set to ``None``.
        But a number of fields can be carried over from ``self``.
        By default, this assumes that the owner of the derived dataset is the same
        as the owner of the original.
        This can be customized with the ``keep`` argument.

        Parameters
        ----------
        keep:
            Fields to copy over to the derived dataset.

        Returns
        -------
        :
            A new derived dataset.

        Raises
        ------
        ValueError
            If ``self`` has no PID.
            The derived dataset requires a PID in order to link back to ``self``.
        """
        if self.pid is None:
            raise ValueError("Cannot make a derived datasets because self.pid is None.")
        return Dataset(
            type=DatasetType.DERIVED,
            input_datasets=[self.pid],
            creation_time=datetime.now(tz=timezone.utc),
            **{name: getattr(self, name) for name in keep},
        )

    def replace_files(self, *files: File) -> Dataset:
        """Return a new dataset with replaced files.

        For each argument, if the input dataset has a file with the
        same remote path, that file is replaced.
        Otherwise, a new file is added.
        Other existing files are kept in the returned dataset.

        Parameters
        ----------
        files:
            New files for the dataset.

        Returns
        -------
        :
            A new dataset with given files.
        """

        def new_or_old(old: File) -> File:
            for new in files:
                if old.remote_path == new.remote_path:
                    return new
            return old

        return self.replace(
            _orig_datablocks=[
                dataclasses.replace(dblock, init_files=map(new_or_old, dblock.files))
                for dblock in self._orig_datablocks
            ]
        )

    def add_orig_datablock(self, *, checksum_algorithm: Optional[str]) -> OrigDatablock:
        """Append a new orig datablock to the list of orig datablocks.

        Parameters
        ----------
        checksum_algorithm:
            Use this algorithm to compute checksums of files
            associated with this datablock.

        Returns
        -------
        :
            The newly added datablock.
        """
        dblock = OrigDatablock(
            checksum_algorithm=checksum_algorithm, _dataset_id=self.pid
        )
        self._orig_datablocks.append(dblock)
        return dblock

    def _lookup_orig_datablock(self, id_: str) -> OrigDatablock:
        try:
            return next(db for db in self._orig_datablocks if db.datablock_id == id_)
        except StopIteration:
            raise KeyError(f"No OrigDatablock with id {id_}") from None

    def _get_or_add_orig_datablock(self, key: Union[int, str, PID]) -> OrigDatablock:
        if isinstance(key, PID):
            key = str(PID)
        if isinstance(key, str):
            return self._lookup_orig_datablock(key)
        # The 0th datablock is implicitly always there and created on demand.
        if key in (0, -1) and not self._orig_datablocks:
            return self.add_orig_datablock(
                checksum_algorithm=self._default_checksum_algorithm
            )
        return self._orig_datablocks[key]

    def make_upload_model(self) -> Union[UploadDerivedDataset, UploadRawDataset]:
        """Construct a SciCat upload model from self."""
        model: Union[Type[UploadRawDataset], Type[UploadDerivedDataset]] = (
            UploadRawDataset if self.type == DatasetType.RAW else UploadDerivedDataset
        )
        # Datablocks are not included here because they are handled separately
        # by make_datablock_upload_models and their own endpoints.
        special = ("relationships", "techniques")
        return model(
            numberOfFiles=self.number_of_files,
            numberOfFilesArchived=self.number_of_files_archived,
            size=self.size,
            packedSize=self.packed_size,
            scientificMetadata=self._meta or None,
            techniques=convert_user_to_upload_model(  # type: ignore[arg-type]
                self.techniques
            ),
            relationships=convert_user_to_upload_model(  # type: ignore[arg-type]
                self.relationships
            ),
            **{
                field.scicat_name: value
                for field in self.fields()
                if field.name not in special
                and (value := getattr(self, field.name)) is not None
            },
        )

    def make_datablock_upload_models(self) -> DatablockUploadModels:
        """Build models for all contained (orig) datablocks.

        Returns
        -------
        :
            Structure with datablock and orig datablock models.
        """
        if self.number_of_files == 0:
            return DatablockUploadModels(orig_datablocks=None)
        return DatablockUploadModels(
            orig_datablocks=[
                dblock.make_upload_model(self) for dblock in self._orig_datablocks
            ]
        )

    def make_attachment_upload_models(self) -> List[UploadAttachment]:
        """Build models for all registered attachments.

        Raises
        ------
        ValueError
            If ``self.attachments`` is ``None``,
            i.e., the attachments are uninitialized.

        Returns
        -------
        :
            List of attachment models.
        """
        if self.attachments is None:
            raise ValueError(
                "Cannot make upload models for attachments because "
                "the attachments are uninitialized."
            )
        return [a.make_upload_model() for a in self.attachments]

    def validate(self) -> None:
        """Validate the fields of the dataset.

        Raises :class:`pydantic.ValidationError` if validation fails.
        Returns normally if it passes.
        """
        self.make_upload_model()

    def keys(self) -> Iterable[str]:
        """Dict-like keys(names of fields) method.

        Returns
        -------
        :
            Generator of names of all fields corresponding to ``self.type``
            and other fields that are not ``None``.


        .. versionadded:: 23.10.0
        """
        from itertools import chain

        all_fields = set((field.name for field in self.fields()))
        my_fields = set((field.name for field in self.fields(dataset_type=self.type)))
        other_fields = all_fields - my_fields
        invalid_fields = (
            f_name for f_name in other_fields if getattr(self, f_name) is not None
        )

        return chain(my_fields, invalid_fields)

    def values(self) -> Iterable[Any]:
        """Dict-like values(values of fields) method.

        Returns
        -------
        :
            Generator of values of all fields corresponding to ``self.type``
            and other fields that are not ``None``.


        .. versionadded:: 23.10.0
        """
        return (getattr(self, field_name) for field_name in self.keys())

    def items(self) -> Iterable[tuple[str, Any]]:
        """Dict-like items(name and value pairs of fields) method.

        Returns
        -------
        :
            Generator of (Name, Value) pairs of all fields
            corresponding to ``self.type``
            and other fields that are not ``None``.


        .. versionadded:: 23.10.0
        """
        return ((key, getattr(self, key)) for key in self.keys())

    @classmethod
    def _validate_field_name(cls, field_name: str) -> None:
        """Validate ``field_name``.

        If ``field_name`` is a ``name`` of any
        :class:`DatasetBase.Field` objects in ``self.fields()``.

        Parameters
        ----------
        field_name:
            Name of the field to validate.

        Raises
        ------
        KeyError
            If validation fails.
        """
        if field_name not in (field.name for field in cls.fields()):
            raise KeyError(f"{field_name} is not a valid field name.")

    def __getitem__(self, field_name: str) -> Any:
        """Dict-like get-item method.

        Parameters
        ----------
        field_name:
            Name of the field to retrieve.

        Returns
        -------
        :
            Value of the field with the name ``field_name``.

        Raises
        ------
        :
            :class:`KeyError` if ``field_name`` does not mach any names of fields.


        .. versionadded:: 23.10.0
        """
        self._validate_field_name(field_name)
        return getattr(self, field_name)

    def __setitem__(self, field_name: str, field_value: Any) -> None:
        """Dict-like set-item method.

        Set the value of the field with name ``field_name`` as ``field_value``.

        Parameters
        ----------
        field_name:
            Name of the field to set.

        field_value:
            Value of the field to set.

        Raises
        ------
        :
            :class:`KeyError` if ``field_name`` does not mach any names of fields.


        .. versionadded:: 23.10.0
        """
        self._validate_field_name(field_name)
        setattr(self, field_name, field_value)


@dataclasses.dataclass
class DatablockUploadModels:
    """Pydantic models for (orig) datablocks."""

    # TODO
    # datablocks: Optional[List[UploadDatablock]]
    orig_datablocks: Optional[List[UploadOrigDatablock]]
    """Orig datablocks"""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

from contextlib import contextmanager
from pathlib import Path
from typing import Generator

import filelock


class FileCounter:
    """Atomic counter based on a file.

    Maintains a file with the count and a separate lock file to
    synchronize access to the counter.
    This is useful in contexts where shared variables between threads or
    processes are not (easily) accessible.
    For instance, when using ``pytest-xdist``, this counter can be used to synchronize
    access to some shared resource between workers.

    The ``increment`` and ``decrement`` functions are intended
    to be used as context managers.
    This allows holding the lock past the increment/decrement and
    synchronizing further operations.
    For example:

    .. code-block:: python

        counter = FileCounter(Path("counter"))
        with counter.increment() as count:
            # do some work that requires the lock
        # lock gets released here
    """

    def __init__(self, path: Path) -> None:
        self._path = path
        self._lock = filelock.FileLock(str(self._path) + ".lock")

    @contextmanager
    def increment(self) -> Generator[int, None, None]:
        with self._lock:
            if self._path.exists():
                count = self._read() + 1
            else:
                count = 1
            self._write(count)
            yield count

    @contextmanager
    def decrement(self) -> Generator[int, None, None]:
        with self._lock:
            count = self._read() - 1
            if count < 0:
                raise RuntimeError(
                    f"Broken file counter, got a file with a count of {count-1}."
                )
            if count == 0:
                self._path.unlink()
            else:
                self._write(count)
            yield count

    def _read(self) -> int:
        with open(self._path, "r") as f:
            c = int(f.read())
            return c

    def _write(self, count: int) -> None:
        with open(self._path, "w") as f:
            f.write(str(count))


class NullCounter:
    """Similar to FileCounter but does not use files, a lock, or count.

    ``increment`` always returns 1 and ``decrement`` always returns 0.
    There is no lock.

    This class can be used similarly to :func:`contextlib.null_context`.
    That is, it allows for transparent handling of conditional counting and locking.
    For example

    .. code-block:: python

        def make_counter():
            if some_condition:
                return FileCounter(path)
            return NullCounter()

        counter = make_counter()
        with counter.increment() as count:
            # Do some work that requires the lock if some_condition is true.
    """

    @contextmanager
    def increment(self) -> Generator[int, None, None]:
        yield 1

    @contextmanager
    def decrement(self) -> Generator[int, None, None]:
        yield 0
